﻿using SqlUtils;
using System;
using System.Collections.Generic;
using System.Configuration;
using System.Data;
using System.Data.SqlClient;
using System.IO;
using WebReports.Api.Scheduler;
using System.Reflection;

namespace ExagoExtensibility
{
    class SchedulerQueue
    {
        #region SchedulerQueue Initialization 
        //Setup Logging
        private static bool writeLog = false;

        //Setup Database
        private static string connectionString = ConfigurationManager.ConnectionStrings["SchedulerQueueDatabaseConnection"].ConnectionString;
        private static DbObjectFactory.DataSourceType db = DbObjectFactory.GetDataSourceType(ConfigurationManager.ConnectionStrings["SchedulerQueueDatabaseConnection"].ProviderName);

        //Setup Exago Settings
        private static bool keepReportsUpToDate = true;
        private static int maxExecutionMinutes = 1440;


        /** 
          * Initializes the queue to setup logging and settings
        */
        static SchedulerQueue()
        {
            //Set logging settings from AppSettings if available
            Boolean.TryParse(ConfigurationManager.AppSettings["sqWriteLog"], out writeLog);

            //Set Exago settings from AppSettings if available
            Boolean.TryParse(ConfigurationManager.AppSettings["sqUpdateReports"], out keepReportsUpToDate);
            Int32.TryParse(ConfigurationManager.AppSettings["sqMaxExecutionMinutes"], out maxExecutionMinutes);
        }

        #endregion
        #region Start Service 
        /** 
		 * Method called when starting a Scheduler Service.
		 * @param servicename has format MachineName:Port.
		 * Sets any stuck jobs in running state [status]=1 to ready [status]=0.
		 */
        public static void Start(string serviceName)
        {
            Utilities.Log(String.Format("Start service {0}", serviceName), writeLog);

            //Refresh Jobs that were stuck running on this service or if they exceed the max execution time
            DbSqlHandler sql = new DbSqlHandler(db, connectionString);
            sql.Stmt = String.Format(
                @"UPDATE {0} " +
                    "SET status = @statusReady, serviceName = NULL, currentExecuteStart = NULL " +
                    "WHERE status = @statusRunning " +
                    "AND ( " +
                        "serviceName = @serviceName " +
                        "OR " +
                        "DATEDIFF(MINUTE, currentExecuteStart, GETUTCDATE()) > @maxExecutionMinutes " +
                        ") " +
                    "AND isDeleted = @false "
                , DbTables.SchedulerJobs);

            sql.SqlParams.Add(new SqlParameter("@statusReady", ScheduleJobStatus.Ready));
            sql.SqlParams.Add(new SqlParameter("@statusRunning", ScheduleJobStatus.Running));
            sql.SqlParams.Add(DbObjectFactory.GetParameter(db, "@false", false, DbType.Boolean));
            sql.SqlParams.Add(new SqlParameter("@serviceName", serviceName));
            sql.SqlParams.Add(new SqlParameter("@maxExecutionMinutes", maxExecutionMinutes));

            sql.ExecuteNQCmd();
        }
        #endregion
        #region GetJobList
        /** 
		 * Method called from the Exago UI to populate the jobs in the Schedule Manager.
		 * @param viewLevel is the schedulemanagerviewlevel setting taken from the active exago session's configuration. Three possible values: all, company and user.
		 * @param companyId is the Parameter to represent the active users' company 
         * @param userId is the Parameter to represent the active users' individual id
		 * @return is a string array of JobListXMl, taken from each scheduled job.
		 */
        public static string[] GetJobList(string viewLevel, string companyId, string userId)
        {
            Utilities.Log(String.Format("GetJobList {0}", viewLevel), writeLog);

            List<string> jobXmlList = new List<string>();
            DbSqlHandler sql = new DbSqlHandler(db, connectionString);
            switch (viewLevel)
            {
                case "company":
                    sql.Stmt = String.Format(
                        @"SELECT jobListXML FROM {0} " +
                            "WHERE createdByCompanyId = @companyId " +
                            "AND  isDeleted = @false"
                            , DbTables.SchedulerJobs);
                    break;
                case "user":
                    sql.Stmt = String.Format(
                        @"SELECT jobListXML FROM {0} " +
                            "WHERE createdByUserId = @userId " +
                            "AND createdByCompanyId = @companyId " +
                            "AND  isDeleted = @false"
                            , DbTables.SchedulerJobs);
                    break;
                default:
                    sql.Stmt = String.Format(
                        @"SELECT jobListXML FROM {0} " +
                            "WHERE isDeleted = @false"
                            , DbTables.SchedulerJobs);
                    break;
            }
            sql.SqlParams.Add(DbObjectFactory.GetParameter(db, "@false", false, DbType.Boolean));
            sql.SqlParams.Add(new SqlParameter("@viewLevel", viewLevel));
            sql.SqlParams.Add(new SqlParameter("@companyId", companyId));
            sql.SqlParams.Add(new SqlParameter("@userId", userId));
            DataSet ds = sql.ExecuteReadCmd();
            foreach (DataRow dr in ds.Tables[0].Rows)
            {
                jobXmlList.Add(dr.ItemArray[0].ToString());
            }
            return jobXmlList.ToArray();
        }
        #endregion
        #region GetNextExecuteJob
        /** 
		 * Method called from the scheduler services to return the next job to execute. 
		 * Schedules periodically query the queue for job assignments. 
         * Searches the database for the next unassigned jobs by checking that servicename is null and status is Ready [0]
		 * @param servicename has format MachineName:Port.
		 * @return the next job to be exected and if there are none, return null. 	 
		 */
        public static string GetNextExecuteJob(string serviceName)
        {
            Utilities.Log(String.Format("GetNextExecuteJob {0}", serviceName), writeLog);

            DbSqlHandler sql = new DbSqlHandler(db, connectionString);

            sql.Stmt = string.Format(
                @"DECLARE @jobId VARCHAR(36)" +
                "SELECT TOP 1 @jobId = jobId " +
                        "FROM {0} " +
                        "WHERE nextExecuteDate <= GETUTCDATE() " +
                        "AND status = @statusReady " +
                        "AND serviceName IS NULL " +
                        "AND isDeleted = @false " +
                        "ORDER BY nextExecuteDate " +
               "UPDATE {0} SET " +
                        "serviceName = @serviceName, status = @statusRunning, currentExecuteStart = @currentExecuteStart " +
                        "WHERE jobId = @jobId AND status = @statusReady " +
               "SELECT jobXML " +
                        "FROM {0} " +
                        "WHERE jobId = @jobId", DbTables.SchedulerJobs);

            sql.SqlParams.Add(new SqlParameter("@currentExecuteStart", DateTime.UtcNow));
            sql.SqlParams.Add(new SqlParameter("@statusReady", ScheduleJobStatus.Ready));
            sql.SqlParams.Add(new SqlParameter("@statusRunning", ScheduleJobStatus.Running));
            sql.SqlParams.Add(DbObjectFactory.GetParameter(db, "@false", false, DbType.Boolean));
            sql.SqlParams.Add(new SqlParameter("@serviceName", serviceName));

            object nextExJob = sql.ExecuteScalarCmd();
            return nextExJob?.ToString();
        }
        #endregion
        #region SaveJob
        /** 
		 * Method called from both the scheduler services and the Exago UI to save schedule jobs.
		 * This method is called when a schedule is added, updated, completed, deleted, etc.
		 * @param jobXml is the XML representation of the scheduled job. 
		 */
        public static void SaveJob(string jobXml)
        {
            Utilities.Log("SaveJob", writeLog);

            //determine inf the job already exists or is new
            QueueApiJob job = QueueApi.GetJob(jobXml);
            DbSqlHandler sql = new DbSqlHandler(db, connectionString);
            sql.Stmt = String.Format(
                @"SELECT COUNT(*) FROM {0} WHERE jobId = @jobId", DbTables.SchedulerJobs);

            sql.SqlParams.Add(new SqlParameter("@jobId", job.JobId));
            int jobCount = (int)sql.ExecuteScalarCmd();
            if (jobCount == 0)
            {
                //If job is new call SaveNewJob() to insert a new db record
                SaveNewJob(jobXml, job);
            }
            else
            {
                //Update existing job record in the database
                SaveExistingJob(jobXml, job, job.Job.JobInfo.NextExecuteDate.Year == 1);
            }

        }

        /** 
        * Method called to insert a new row in the schedulerjob table when a new schedule is created
        * @param jobXml is the XML representation of the scheduled job. 
        * @param job is the QueueApiJob object derived from the jobXml.
        */
        public static void SaveNewJob(string jobXml, QueueApiJob job)
        {
            Utilities.Log(String.Format("SaveNewJob - name {0}", job.ScheduleName), writeLog);
            DbSqlHandler sql = new DbSqlHandler(db, connectionString);
            string userId = job.Job.JobInfo.UserId;
            string companyId = job.Job.JobInfo.CompanyId;

            sql.Stmt = String.Format(
                        @"INSERT INTO {0}(jobId, nextExecuteDate, status, createdByCompanyId, createdByUserId, lastModifiedByCompanyId, " +
                            "lastModifiedByUserId, isDeleted, scheduleName, reportName, reportId, reportXml, jobXml, jobListXml) " +
                        "VALUES( @jobId, @nextExecuteDate, @status,  @companyId, @userId, @companyId, @userId, @false, @scheduleName, " +
                            "@reportName, @reportId, @reportXML, @jobXML, @jobListXML)"
                        , DbTables.SchedulerJobs);

            sql.SqlParams.Add(DbObjectFactory.GetParameter(db, "@false", false, DbType.Boolean));
            sql.SqlParams.Add(new SqlParameter("@jobId", job.JobId));
            sql.SqlParams.Add(new SqlParameter("@nextExecuteDate", job.Job.JobInfo.NextExecuteInstant.ToDateTimeUtc()));
            sql.SqlParams.Add(new SqlParameter("@status", job.Status));
            sql.SqlParams.Add(new SqlParameter("@companyId", companyId));
            sql.SqlParams.Add(new SqlParameter("@userId", userId));
            sql.SqlParams.Add(new SqlParameter("@scheduleName", job.ScheduleName));
            sql.SqlParams.Add(new SqlParameter("@reportName", job.ReportName));
            sql.SqlParams.Add(new SqlParameter("@reportId", job.ReportId));
            sql.SqlParams.Add(new SqlParameter("@reportXML", job.ReportXml));
            sql.SqlParams.Add(new SqlParameter("@jobXML", jobXml));
            sql.SqlParams.Add(new SqlParameter("@jobListXML", job.JobListXml));

            sql.ExecuteNQCmd();
        }
        /** 
       * Method called to update the db record of an existing job.
       * Utilized when a user changes a schedule in the scheduler manager, when a schedule compelete running, or when a schedule is deleted
       * @param jobXml is the XML representation of the scheduled job. 
       * @param job is the QueueApiJob object derived from the jobXml.
       * @param isComplete is whether the schedule will run again
        */
        public static void SaveExistingJob(string jobXml, QueueApiJob job, bool isComplete)
        {
            Utilities.Log(String.Format("SaveExistingJob - name {0} status {1}", job.ScheduleName, job.Status), writeLog);
            DbSqlHandler sql = new DbSqlHandler(db, connectionString);
            string userId = job.Job.JobInfo.UserId;
            string companyId = job.Job.JobInfo.CompanyId;
            sql.Stmt = String.Format(
                        @"UPDATE {0} " +
                            "SET nextExecuteDate=@nextExecuteDate, status=@status, lastModifiedByCompanyId=@companyId, lastModifiedByUserId=@userId, " +
                                "scheduleName=@scheduleName, reportName=@reportName, reportId=@reportId, reportXml=@reportXML, jobXml=@jobXML, jobListXml=@jobListXML, currentExecuteStart=NULL " +
                        "WHERE jobId=@jobId "
                        , DbTables.SchedulerJobs);

            //Handles Resetting the Service Name if the job is ready to be re-executed
            if (job.Status == JobStatus.Ready)
            {
                sql.Stmt += String.Format(
                        @"UPDATE {0} " +
                            "SET serviceName=NULL " +
                        "WHERE jobId=@jobId "
                        , DbTables.SchedulerJobs);
                sql.SqlParams.Add(new SqlParameter("@serviceName", DBNull.Value));
            }

            //Handles Resetting the NextExecuteDate if the job is in a completed state
            if (isComplete)
            {
                sql.SqlParams.Add(new SqlParameter("@nextExecuteDate", DBNull.Value));
            }
            else
            {
                sql.SqlParams.Add(new SqlParameter("@nextExecuteDate", job.Job.JobInfo.NextExecuteInstant.ToDateTimeUtc()));
            }

            sql.SqlParams.Add(new SqlParameter("@jobId", job.JobId));
            sql.SqlParams.Add(new SqlParameter("@status", job.Status));
            sql.SqlParams.Add(new SqlParameter("@companyId", companyId));
            sql.SqlParams.Add(new SqlParameter("@userId", userId));
            sql.SqlParams.Add(new SqlParameter("@scheduleName", job.ScheduleName));
            sql.SqlParams.Add(new SqlParameter("@reportName", job.ReportName));
            sql.SqlParams.Add(new SqlParameter("@reportId", job.ReportId));
            sql.SqlParams.Add(new SqlParameter("@reportXML", job.ReportXml));
            sql.SqlParams.Add(new SqlParameter("@jobXML", jobXml));
            sql.SqlParams.Add(new SqlParameter("@jobListXML", job.JobListXml));

            sql.ExecuteNQCmd();
        }

        #endregion
        #region GetJobData
        /** 
		 * Method called when a user edits a schedule in the Scheduler Manager to get the full job XML data for a job.
		 * @param jobId is the unique ID (GUID) for a scheduled job. 
		 * @returns jobXML for the job requested by the Schedule Manager. 
		 */


        public static string GetJobData(string jobId)
        {
            Utilities.Log(String.Format("GetJobData", jobId), writeLog);

            DbSqlHandler sql = new DbSqlHandler(db, connectionString);
            sql.Stmt = String.Format(
                @"SELECT jobXml FROM {0} WHERE jobId = @jobId"
                , DbTables.SchedulerJobs);

            sql.SqlParams.Add(new SqlParameter("@jobId", jobId));
            return sql.ExecuteScalarCmd()?.ToString();
        }
        #endregion
        #region DeleteReport 
        //Delete Schedules upon Report Deletion should be set to True in the Admin console for this method to get hit 
        /** 
		 * Method called from the Exago UI when a report is deleted.
		 * Does a 'soft delete' by modifying a boolean 'isDeleted' column in the database. 
		 * This indicates to the Scheduler Queue that the scheduled job should be ignored. '
         * @param reportId Guid used to uniquely identify schedules using the deleted report. 
		 * Method is only hit if setting: Delete Schedules upon Report Deletion is set to True in the Admin console.
		*/
        public static void DeleteReport(string reportId)
        {
            Utilities.Log(String.Format("DeleteReport {0}", reportId), writeLog);

            DbSqlHandler sql = new DbSqlHandler(db, connectionString);
            sql.Stmt = String.Format(
                @"UPDATE {0} " +
                    "SET isDeleted= @true, status= @status " +
                    "WHERE reportId=@reportId"
                    , DbTables.SchedulerJobs);

            sql.SqlParams.Add(new SqlParameter("@status", JobStatus.Deleted));
            sql.SqlParams.Add(DbObjectFactory.GetParameter(db, "@true", true, DbType.Boolean));
            sql.SqlParams.Add(new SqlParameter("@reportId", reportId));

            sql.ExecuteNQCmd();
            return;

        }
        #endregion
        #region RenameReport
        /** 
		 * Method called from the Exago UI when a report is renamed.
		 * @param reportId Guid used to uniquely identify the report being renamed. 
		 * @param reportName is the new report's new name. 
		 */
        public static void RenameReport(string reportId, string reportName)
        {
            Utilities.Log(String.Format("RenameReport {0}", reportId), writeLog);

            if (!keepReportsUpToDate)
                return;

            DbSqlHandler sql = new DbSqlHandler(db, connectionString);
            sql.Stmt = String.Format(
                @"UPDATE {0} " +
                    "SET reportName=@reportName WHERE reportId=@reportId"
                , DbTables.SchedulerJobs);
            sql.SqlParams.Add(new SqlParameter("@reportId", reportId));
            sql.SqlParams.Add(new SqlParameter("@reportName", reportName));

            sql.ExecuteNQCmd();
        }
        #endregion
        #region UpdateReport
        /** 
		 * Method called from the Exago UI when a report is updated. 
		 * Updates report contents for any scheduled jobs that contain the specific report id.
         * @param reportId Guid used to uniquely identify schedules using the updated report. 
		 * @param reportXml is the updated XML representation of the report. 
		 */
        public static void UpdateReport(string reportId, string reportXml)
        {
            Utilities.Log(String.Format("UpdateReport {0}", reportId), writeLog);

            if (!keepReportsUpToDate)
                return;
            DbSqlHandler sql = new DbSqlHandler(db, connectionString);
            sql.Stmt = String.Format(
                @"UPDATE {0} " +
                    "SET reportXml=@reportXml " +
                    "WHERE reportId=@reportId"
                , DbTables.SchedulerJobs);
            sql.SqlParams.Add(new SqlParameter("@reportId", reportId));
            sql.SqlParams.Add(new SqlParameter("@reportXml", reportXml));

            sql.ExecuteNQCmd();
        }
        #endregion
        #region Flush 
        /** 
		 * Method called from the Exago UI Scheduler Manager in response to a click on the Flush button.
		 * Soft deletes any jobs that are marked as deleted or completed.  
		 * @param viewLevel is the schedulemanagerviewlevel setting taken from the active exago session's configuration. Three possible values: all, company and user.
		 * @param companyId is the Parameter to represent the active users' company 
         * @param userId is the Parameter to represent the active users' individual id
		 * @return is a string array of JobListXMl, taken from each scheduled job.
		 */
        public static void Flush(string viewLevel, string companyId, string userId)
        {
            Utilities.Log(String.Format("Flush {0}", viewLevel), writeLog);

            DbSqlHandler sql = new DbSqlHandler(db, connectionString);
            switch (viewLevel)
            {
                case "company":
                    sql.Stmt = String.Format(
                        @"UPDATE {0} " +
                            "SET isDeleted= @true " +
                            "WHERE status IN (@statusCompelete, @statusDeleted, @statusRemoved) AND createdByCompanyId = @companyId "
                            , DbTables.SchedulerJobs);
                    break;
                case "user":
                    sql.Stmt = String.Format(
                        @"UPDATE {0} " +
                            "SET isDeleted= @true " +
                            "WHERE status IN (@statusCompelete, @statusDeleted, @statusRemoved) AND createdByUserId = @userId AND createdByCompanyId = @companyId "
                        , DbTables.SchedulerJobs);
                    break;
                default:
                    sql.Stmt = String.Format(
                        @"UPDATE {0} " +
                            "SET isDeleted= @true " +
                            "WHERE Status IN (@statusCompelete, @statusDeleted, @statusRemoved) "
                            , DbTables.SchedulerJobs);
                    break;
            }

            //Reset jobs exceeding the max Execution Time.
            sql.Stmt += String.Format(
                @"UPDATE {0} " +
                    "SET status = @statusReady, serviceName = NULL, currentExecuteStart = NULL " +
                    "WHERE status = @statusRunning " +
                    "AND DATEDIFF(MINUTE, currentExecuteStart, GETUTCDATE()) > @maxExecutionMinutes " +
                    "AND isDeleted = @false "
                , DbTables.SchedulerJobs);

            sql.SqlParams.Add(new SqlParameter("@statusCompelete", ScheduleJobStatus.Completed));
            sql.SqlParams.Add(new SqlParameter("@statusDeleted", ScheduleJobStatus.Deleted));
            sql.SqlParams.Add(new SqlParameter("@statusRemoved", ScheduleJobStatus.Removed));
            sql.SqlParams.Add(new SqlParameter("@statusReady", ScheduleJobStatus.Ready));
            sql.SqlParams.Add(new SqlParameter("@statusRunning", ScheduleJobStatus.Running));
            sql.SqlParams.Add(new SqlParameter("@maxExecutionMinutes", maxExecutionMinutes));
            sql.SqlParams.Add(DbObjectFactory.GetParameter(db, "@true", true, DbType.Boolean));
            sql.SqlParams.Add(DbObjectFactory.GetParameter(db, "@false", true, DbType.Boolean));
            sql.SqlParams.Add(new SqlParameter("@viewLevel", viewLevel));
            sql.SqlParams.Add(new SqlParameter("@companyId", companyId));
            sql.SqlParams.Add(new SqlParameter("@userId", userId));

            sql.ExecuteNQCmd();
        }
        #endregion
    }
}
